"""
==================================
Base Mesh (:mod:`discretize.base`)
==================================
.. currentmodule:: discretize.base

The ``base`` sub-package houses the fundamental classes for all meshes in ``discretize``.

Base Mesh Class
---------------
.. autosummary::
  :toctree: generated/

  BaseMesh
  BaseRectangularMesh
  BaseTensorMesh
"""

from discretize.base.base_mesh import BaseMesh, BaseRectangularMesh
from discretize.base.base_tensor_mesh import BaseTensorMesh
